/*
 * $Source: r:/source/msgapi/qqlink/RCS/geconfig.c,v $
 * $Revision: 1.5 $
 * $Date: 2000/12/10 00:56:35 $
 * $Locker:  $
 *
 *	Read GEcho Area Definitions (written by Alex Woick)
 *
 * $Log: geconfig.c,v $
 * Revision 1.5  2000/12/10 00:56:35  vitus
 * - printf() durch Verbose() ersetzt
 *
 * Revision 1.4  1999/07/01 23:46:04  vitus
 * - includes qqlink.h, local declarations removed
 *
 * Revision 1.3  1998/12/19 00:47:28  vitus
 * - changed directory of common header files
 *
 * Revision 1.2  1998/11/12 03:00:46  vitus
 * - RCS variables
 * - changed C style to match rest of program
 */
static char const vcid[]="$Id: geconfig.c,v 1.5 2000/12/10 00:56:35 vitus Exp $";

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <share.h>

#define O_DENYWRITE	SH_DENYWR


typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

#include "../../Lib/defines.h"
#define HFILE	int				/* keep compiler happy */
#include "../../Lib/verbose.h"


#include "gestruct.h"
#include "qqlink.h"




SETUP_GE	sys;
AREAFILE_HDR	AreaHdr;
AREAFILE_GE	Area;
AREAFILE_GEX	* AreaB, * AreaP;
CONNECTION	Connection[MAXCONNECTIONS],	* ConnectionP;

int		SetupGE, AreaFileGE, AreaFileGEX;
word		areas, arearecsize, connectionsize;




PUBLIC int
do_gechocfg(char *gepath)
{
    int counter, result; /* records, node; */
    long temp;
    char s[_MAX_PATH];


    assert( sizeof(SETUP_GE) == 5764 );		/* bad structure packing */

    /*
     * Opening and reading SETUP.GE
     */

    sprintf(s, "%s\\SETUP.GE", gepath);
    SetupGE = open(s, O_RDONLY|O_DENYWRITE|O_BINARY);
    if( SetupGE == -1 )
    {
	Verbose(0,"QQLINK", "Unable to open %s - errno %u", s, errno);
	return -1;
    }
    result = read(SetupGE, &sys, sizeof(SETUP_GE));
    close(SetupGE);
    Verbose(1,"QQLINK", "gelesen: %d, erwartet: %d", result, sizeof(SETUP_GE));

    if( result < sizeof(SETUP_GE) )
    {
	Verbose(0,"QQLINK", "Error reading %s - errno %u", s, errno);
	return -1;
    }

    if( sys.sysrev != GE_THISREV )
    {
	Verbose(0,"QQLINK", "GEcho Config: System file revision level mismatch");
	return -1;
    }

    Verbose(1,"QQLINK", "System file created by GEcho %u.%02u",
	    sys.version_major, sys.version_minor);


    /*
     *  Opening AREAFILE.GE and checking the header
     */
    sprintf(s, "%s\\AREAFILE.GE", gepath);
    AreaFileGE = open(s, O_RDWR|O_DENYWRITE|O_BINARY);
    if( AreaFileGE == -1 )
    {
	Verbose(0,"QQLINK", "Unable to open %s - errno %u", s, errno);
	return -1;
    }

    result = read(AreaFileGE, &AreaHdr, sizeof(AREAFILE_HDR));
    if( result < sizeof(AREAFILE_HDR) )
    {
	Verbose(0,"QQLINK", "Error reading GEcho AREAFILE header - errno %u", errno);
	close(AreaFileGE);
	return -1;
    }

    if( AreaHdr.hdrsize < sizeof(AREAFILE_HDR) ||
	AreaHdr.recsize < sizeof(AREAFILE_GE) )
    {
	Verbose(0,"QQLINK", "Incompatible GEcho AREAFILE record size");
	close(AreaFileGE);
	return -1;
    }

    connectionsize = AreaHdr.maxconnections * sizeof(CONNECTION);
    arearecsize = AreaHdr.recsize + connectionsize;

    /*
     *  Reading AREAFILE.GE using AREAFILE.GEX index file
     */

    sprintf(s, "%s\\AREAFILE.GEX", gepath);
    AreaFileGEX = open(s, O_RDWR|O_DENYWRITE|O_BINARY);
    if( AreaFileGEX == -1 )
    {
	Verbose(0,"QQLINK", "Unable to open %s - errno", s, errno);
	close(AreaFileGE);
	return -1;
    }

    temp = filelength(AreaFileGEX) / sizeof(AREAFILE_GEX);
    if( temp > MAXAREAS )
    {
	Verbose(0,"QQLINK", "GEcho: Index file too large");
	close(AreaFileGE);
	close(AreaFileGEX);
	return -1;
    }
    areas = (word)temp;

    AreaB = malloc(areas * sizeof(AREAFILE_GEX));
    if( AreaB == NULL )
    {
	Verbose(0,"QQLINK", "Insufficient memory");
	close(AreaFileGE);
	close(AreaFileGEX);
	return -1;
    }

    read(AreaFileGEX, AreaB, areas * sizeof(AREAFILE_GEX));
    close(AreaFileGEX);

    for( AreaP = AreaB, counter = 1; counter <= areas; counter++, AreaP++)
    {
	lseek(AreaFileGE, AreaP->offset, SEEK_SET);
	result = read(AreaFileGE, &Area, sizeof(AREAFILE_GE));
	if( result == sizeof(AREAFILE_GE) )
	{
	    if( Area.areaformat==FORMAT_SQUISH )
	    {
		add_area(Area.name, Area.path, 1, Area.areatype==NETMAIL);
	    }
	}
    }
    close(AreaFileGE);
    free(AreaB);
    return 0;
}
