/*
 * $Source: r:/source/msgapi/qqlink/RCS/qconfig.c,v $
 * $Revision: 1.8 $
 * $Date: 2000/05/29 00:51:05 $
 * $Locker:  $
 *
 *	QQLink 1.24.vx - 
 *
 * $Log: qconfig.c,v $
 * Revision 1.8  2000/05/29 00:51:05  vitus
 * - parse_adotb: ignore lines starting with "P<blank>"
 *
 * Revision 1.7  2000/05/29 00:33:54  vitus
 * - parse_adotb: increased buffer space to _MAX_PATH + MAX_TAGLEN
 *
 * Revision 1.6  1999/07/01 23:44:27  vitus
 * - parse_adotb: read area tags into MAX_TAGLEN buffer
 * - includes qqlink.h, local declarations removed
 *
 * Revision 1.5  1998/12/19 00:45:10  vitus
 * - changed directory of common header files
 *
 * Revision 1.4  1998/11/13 02:15:49  vitus
 * - added A.Woiks' do_max3areadat(char *fn)
 * - calls A.Woiks' do_gechocfg(char *gepath)
 * - switched to Scott's SQDEV and MSTRUCTS, minor changes
 *
 * Revision 1.3  1998/06/05 09:20:15  vitus
 * - replaced printf() with Verbose()
 *
 * Revision 1.2  1997/09/16 10:23:40  vitus
 * - beautified
 * - recognizes LocalArea
 */
static char const vcid[]="$Id: qconfig.c,v 1.8 2000/05/29 00:51:05 vitus Exp $";

/**
***  This file is part of the source to QQLink, a FTS-9 reply linker for
*** use with the Squish(tm) mail processor.
*** Copyright (c) 1993 Josh Parsons. All rights reserved, however:
***
***         1) You are permitted to freely distribute this source,
***           provided you do not remove this copyright notice, and
***           provided you do not charge a fee, for the service of
***           distributing this material, over and above the value
***           of disks, phone tolls, etc. incurred in such distribution.
***        2) You are permitted to use executables produced from this
***           source in the manner described in QQLink.DOC, which should
***           accompany this file.
***        3) You are permitted to use this source in your own programs,
***           provided that your derivative work carries an
***           acknowledgement that portions of its source are derived
***           from QQLink, by Josh Parsons.
***
***        NB: This source carries no warrantee, implicit or explicit
***            of fitness for any particular task.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "max.h"				/* from Maximus distribution! */
#include "MsgApi.h"

#include "../../Lib/defines.h"
#define HFILE	int				/* keep compiler happy */
#include "../../Lib/verbose.h"

#include "config.h"
#include "qqlink.h"






PUBLIC char	* squishcfg_fname = "Squish.CFG";
PUBLIC char	* areasbbs_fname = NULL;
PUBLIC char	* max3areadat_fname = NULL;
PUBLIC char	* gepath = NULL;
PUBLIC char	* echolog_fname = NULL;




#define PARCMP(x,y)	strnicmp(x,y,strlen(y))
#define PARGET(x,y)	((x)+strlen(y))




PRIVATE int	flgseen[CFG_maxargs];




PRIVATE char *
flgget(int argc,char **argv,const char *flg)
{
    int		i;
    char *	ret = NULL;

    if( flg[strlen(flg)-1] == ':'  ||  *flg == '-' )
    {
	for( i = 0; i < argc; ++i )
	    if( !PARCMP(argv[i], flg) )
	    {
		++flgseen[i];
		ret = PARGET(argv[i], flg);
	    }
    }
    else
	for( i = 0; i < argc; ++i )
	    if( !stricmp(argv[i], flg) )
	    {
		++flgseen[i];
		ret = argv[i];
	    }
    return ret;
}




PRIVATE void
flgclr(void)
{
    int i;

    for( i = 0; i < CFG_maxargs; ++i )
	flgseen[i] = 0;
}




PRIVATE int
flgchk(int argc,char **argv)
{
    int ret = 0, i;

    for( i = 0; i < argc; ++i )
	if( flgseen[i] == 0 )
	    ret |= CFG_error("Unrecognised flag \"%s\".", argv[i]);
    return ret;
}




PRIVATE int
do_squish_cfg(int argc,char **argv,int flgc,char **flgv)
{
    if( argc < 1 )
	return 0;			/* should never happen, but anyway... */
    flgclr();
    if( !stricmp(*argv,"EchoArea") )
    {
	if( argc < 3 )
	    return CFG_error("Wrong number of arguments to %s.", *argv );
	if( !flgget(flgc, flgv, "-0")  ||  link_passthru )
	    add_area(argv[1], argv[2], flgget(flgc, flgv, "-$") != NULL, 0);
    }
    else if( !stricmp(*argv,"LocalArea") )
    {
	if( argc < 3 )
	    return CFG_error("Wrong number of arguments to %s.", *argv );
	if( !flgget(flgc, flgv, "-0")  ||  link_passthru )
	    add_area(argv[1], argv[2], flgget(flgc, flgv, "-$") != NULL, 0);
    }
    else if( !stricmp(*argv, "NetArea") )
    {
	if( argc < 3 )
	    return CFG_error("Wrong number of arguments to %s.", *argv);
	if( !flgget(flgc, flgv, "-0") || link_passthru )
	    add_area(argv[1], argv[2], flgget(flgc, flgv,"-$") != NULL, 1);
    }
    else if( !stricmp(*argv, "AreasBBS") )
    {
	if( argc != 2 )
	    return CFG_error("Wrong number of arguments to %s.", *argv);
	areasbbs_fname = strupr(strdup(argv[1]));
    }
    return 0;
}




PRIVATE int
parse_adotb(const char *fn)
{
    static char	buf[_MAX_PATH+MAX_TAGLEN+20];
    char *	path, * tag;
    int		lines = 0;
    FILE *	fin = fopen(fn, "r");

    if( fin == NULL )
	return -1;
    while( !feof(fin) )
    {
	char *	s;
	int	sq = 0, cmnt = 0;

	fgets(buf, sizeof(buf), fin);
	s = strchr(buf, '\n');
	if( s )
	    *s = 0;

	s = buf;
	if( buf[0] == 'P'  &&  isspace(buf[1]) )
	    continue;				/* ignore: "P <tag> <address>" */

	while( s )
	{
	    if( isspace(*s) )
		;
	    else if( *s == '#')
		;
	    else if( *s == '$' )
		sq = 1;
	    else if( *s == ';' )
		cmnt = 1;
	    else
		break;
	    ++s;
	}
	if( cmnt )
	    continue;

	path = s;
	while( s  &&  !isspace(*s) )
	    ++s;
	*(s++) = 0;				/* end 'path' with \000 */

	while( s  &&  isspace(*s) )
	    ++s;				/* ignore white space */

	tag = s;
	while( s  &&  !isspace(*s) )
	    ++s;
	*(s++) = 0;				/* end 'tag' with \000 */

	if( *tag  &&  *path )
	{
	    /* Avoid origin/sysop line */
	    if( lines++ )
		add_area(tag, path, sq,0);
	}
    }
    fclose(fin);
    return 0;
}






/*# ----------------------------------------------------------------------
 * CALL
 *	do_max3areadat(fn)
 * PARAMETER
 *	fn		Filename of Maximus area.dat
 *
 * RETURNS
 *	0		OK
 *	/0		error
 *
 * DESCRIPTION
 *	Parse Maximus 3.0 area.dat.
 *
 * REMARKS
 */
PRIVATE int
do_max3areadat(char *fn)
{
    FILE	* marea;
    MAH		mah;
    dword	id;

    marea = fopen(fn, "rb");
    if( marea == NULL )
	return -1;

    fread(&id, sizeof(id), 1, marea);
    if( id != MAREA_ID )
    {
	fclose(marea);
	return -1;
    }

    fseek(marea, ADATA_START, SEEK_SET);
    while( !feof(marea) )
    {
	long numread = fread(&(mah.ma), sizeof(MAREA), 1, marea);

	fseek(marea, mah.ma.cbArea-numread*sizeof(MAREA), SEEK_CUR);
	mah.pov = malloc(sizeof(OVERRIDE)*mah.ma.num_override);
	numread = fread(mah.pov, sizeof(OVERRIDE), mah.ma.num_override, marea);
	mah.heap = malloc(mah.ma.cbHeap);
	numread = fread(mah.heap, mah.ma.cbHeap, 1, marea);

	if( (mah.ma.attribs & (MA_DIVBEGIN|MA_DIVEND)) == 0 )
	{
	    add_area(MAS(mah, echo_tag), MAS(mah,path),
		     mah.ma.type == MSGTYPE_SQUISH, (mah.ma.attribs & MA_NET) > 0);
	}

	free(mah.heap);
	free(mah.pov);
    }
    fclose(marea);
    return 0;
}




PUBLIC void
q_config(void)
{
    if( CFG_parse(squishcfg_fname, do_squish_cfg) == CFGERR_syntax )
    {
	Verbose(0,"QQLINK", "Error(s) %s config file %s","parsing", squishcfg_fname);
	exit(10);
    }
    if( areasbbs_fname  &&  parse_adotb(areasbbs_fname) )
    {
	Verbose(0,"QQLINK","Error(s) %s config file %s", "reading", areasbbs_fname);
	exit(10);
    }
    if( max3areadat_fname  &&  do_max3areadat(max3areadat_fname) )
    {
	printf("\n\nError(s) %s config file %s\n","reading", max3areadat_fname);
	exit(10);
    }

    if( gepath  &&  do_gechocfg(gepath) )
    {
	printf("\n\nError(s) %s GEcho config in %s\n","reading",gepath);
	exit(10);
    }

    return;
}
