/*
 * $Source: r:/source/msgapi/qqlink/RCS/qqdat.c,v $
 * $Revision: 1.10 $
 * $Date: 2000/11/21 02:13:38 $
 * $Author: vitus $
 *
 *	QQLink 1.24.jx - Linking
 *
 * $Log: qqdat.c,v $
 * Revision 1.10  2000/11/21 02:13:38  vitus
 * - link_qq: now returns negative value when something could not be done,
 *   positive values are "count of linked messages".
 * - commented
 *
 * Revision 1.9  2000/11/17 02:20:37  vitus
 * - moved last_* (detection of unneeded writes) to QQDAT structure.  The
 *   global variable approach is not usable in multithreaded programs...
 * - changed MSGH* to HMSG (as sqdev suggests)
 *
 * Revision 1.8  2000/11/16 23:15:51  vitus
 * - code stores and scans REPLY IDs to even link out-of-order msgs
 *
 * Revision 1.7  1999/07/02 19:28:56  vitus
 * - includes qqlink.h, local declarations removed
 *
 * Revision 1.6  1998/12/19 00:45:31  vitus
 * - changed directory of common header files
 *
 * Revision 1.5  1998/09/02 01:30:57  vitus
 * - corrected MSGID/REPLY kludge extraction
 * - added alternate ID calculations: MD5, CRC32 (used)
 *
 * Revision 1.4  1998/06/05 09:20:28  vitus
 * - replaced printf() with Verbose()
 *
 * Revision 1.3  1997/09/17 01:21:50  vitus
 * - marked all symbols either PRIVATE or PUBLIC
 *
 * Revision 1.2  1997/09/16 10:26:11  vitus
 * - beautified
 */
#pragma strings(readonly)
static char const vcid[]="$Id: qqdat.c,v 1.10 2000/11/21 02:13:38 vitus Exp $";

/**
***  This file is part of the source to QQLink, a FTS-9 reply linker for
*** use with the Squish(tm) mail processor.
*** Copyright (c) 1993 Josh Parsons. All rights reserved, however:
***
***         1) You are permitted to freely distribute this source,
***           provided you do not remove this copyright notice, and
***           provided you do not charge a fee, for the service of
***           distributing this material, over and above the value
***           of disks, phone tolls, etc. incurred in such distribution.
***        2) You are permitted to use executables produced from this
***           source in the manner described in QQLink.DOC, which should
***           accompany this file.
***        3) You are permitted to use this source in your own programs,
***           provided that your derivative work carries an
***           acknowledgement that portions of its source are derived
***           from QQLink, by Josh Parsons.
***
***        NB: This source carries no warrantee, implicit or explicit
***            of fitness for any particular task.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "compiler.h"
#include "msgapi.h"

#include "../../Lib/defines.h"			/* my standard defines */
#define HFILE	int				/* keep compiler happy */
#include "../../Lib/verbose.h"

#include "qqlink.h"
#include "qqdat.h"



/* Define MSGID calculation (0: QQLink, 1: MD5, 2:CRC32).
 * QQLink:	Andreas Wei reported message IDs resulting in incorrect
 *		linking.  Don't use it anymore.
 * MD5:		got many incorrect links using my own TEAMOS2.GER message
 *		base.  Needs larger data buffers?
 * CRC32:	seems to be the way to go
 */
#define MAKEID_METHOD	2



#if MAKEID_METHOD == 1
# include "md5.h"
#endif
#if MAKEID_METHOD == 2
# include "crc32.h"
#endif




PUBLIC int	trunc_qq = 0;






#if MAKEID_METHOD == 0
PRIVATE void
makeid(PMSGID const id, const char *s)
{
    char	b = 0;
    int		ctr = trunc_qq;

    while( *s )
    {
	*id += ((long)tolower(*s)) << b;
 	b = (b+1) % 24;
 	++s;
 	if( (--ctr) == 0 )
	    break;
    }
    if( *id == 0 )
  	*id = 1;		/* ID=0 is used for "no ID found" by getid() */
}
#endif


#if MAKEID_METHOD == 1
PRIVATE void
makeid(PMSGID id,const char *s)
{
    char const *	buf = strlwr(strdup(s));
    int const		size = MIN(strlen(buf),trunc_qq);
    struct MD5Context	ctx;
    unsigned char	digest[16];
    unsigned long const * p = (unsigned long const *)digest;

    MD5Init(&ctx);
    MD5Update(&ctx, buf, size);
    MD5Final(digest, &ctx);
    *id = (((p[0] + p[1]) >> 1)
	   + ((p[2] + p[3]) >> 1));

    if( *id == 0 )
	*id = 1;		/* ID=0 is used for "no ID found" by getid() */
    return;
}
#endif


#if MAKEID_METHOD == 2
PRIVATE void
makeid(PMSGID id,const char *s)
{
    int		ctr = trunc_qq;
    CRC32	crc;        

    crc32Init(&crc);
    while( *s )
    {
	crc32Update(&crc, tolower(*s));
 	++s;
 	if( (--ctr) == 0 )
	    break;
    }
    *id = crc32Value(&crc);

    if( *id == 0 )
	*id = 1;		/* ID=0 is used for "no ID found" by getid() */
    return;
}
#endif




/*# ----------------------------------------------------------------------
 * getid(hmsg, id, oid)
 *
 * PARAMETER
 *	hmsg		handle of opened message
 *	id		room for MSGID based on MSGID kludge
 *	oid		room for MSGID based on REPLY kludge
 *
 * RETURNS
 *	(nothing)
 *
 * DESCRIPTION
 *
 * REMARKS
 *	Kludge-lines may end with \000, \r or \001 (a new kludge-line)!
 */
PRIVATE void
getid(HMSG const hmsg, PMSGID const id, PMSGID const oid)
{
    char	*msgid, *reply;
    char	cbuf[1024];

    MsgReadMsg(hmsg, NULL, 0, 0, NULL, sizeof(cbuf), cbuf);
    msgid = strstr(cbuf, "\001MSGID: ");
    reply = strstr(cbuf, "\001REPLY: ");
    *id = *oid = 0;
    if( msgid )
    {
	char	*s;

	msgid += 8;
	s = strpbrk(msgid, "\r\n\001");
	if( s )
	    *s = 0;				/* end string after kludge */
	makeid(id, msgid);
    }
    if( reply )
    {
	char	*s;

	reply += 8;
	s = strpbrk(reply, "\r\n\001");
	if( s )
	    *s = 0;				/* s.o. */
	makeid(oid, reply);
    }
}




/*# ----------------------------------------------------------------------
 * flushxmsg(qq, i)
 *
 * PARAMETER
 *	qq		data of related .QQL file
 *	i		index in xmsg read cache
 *
 * RETURNS
 *	0		did nothing
 *	1		wrote to disk
 *
 * DESCRIPTION
 *	If the index 'i' contains a valid and modified xmsg write
 *	that xmsg back to disk.
 *
 * REMARKS
 */
PRIVATE int
flushxmsg(PQQDAT const qq, int const i)
{
    int		result = 0;
    HMSG	hmsg;

    if( qq->xmnum[i] < 1  ||  qq->xmflg[i] == 0 )
	return result;				/* Leave the memorial NOW! */
    hmsg = MsgOpenMsg(qq->area, MOPEN_RW, qq->xmnum[i]);
    if( hmsg == NULL )
	Verbose(0,"QQLINK","Internal error flushing message %ld (%d)",qq->xmnum[i],i);
    else
    {
	MsgWriteMsg(hmsg, 0, qq->xmbuf+i, NULL, 0, 0, 0, NULL);
	MsgCloseMsg(hmsg);
	result = 1;
    }
    qq->xmflg[i] = 0;
    return result;
}




/*# ----------------------------------------------------------------------
 * readxmsg(qq, hmsg, msgn)
 *
 * PARAMETER
 *	qq		data of related .QQL file
 *	hmsg		handle of opened message 'msgn'
 *			(if NULL readxmsg() can't read the message)
 *	msgn		message to read from
 *
 * RETURNS
 *	pointer to messgae header information (XMSG)
 *
 * DESCRIPTION
 *	Read and return message header.
 *
 * REMARKS
 */
PRIVATE PXMSG
readxmsg(PQQDAT qq, HMSG hmsg, long const msgn)
{
    int i, j;

    qq->last[0].xnum = msgn;

    /* See if we've got the header in one of our buffers.  If we
     * have it's faster to return that... */

    for( i = 0; i < MAX_XMSG; ++i )
	if( qq->xmnum[i] == msgn )
	{
	    memcpy(&qq->last[0].xmsg, qq->xmbuf+i, sizeof(XMSG));
	    return qq->xmbuf+i;
	}
    if( hmsg == NULL )
	return NULL;


    /* See if there's an unused buffer to load it into */

    for( i = 0; i < MAX_XMSG; ++i )
	if( qq->xmnum[i] == 0 )
	{
	    MsgReadMsg(hmsg, qq->xmbuf+i, 0, 0, NULL, 0, NULL);
	    qq->xmnum[i] = msgn;
	    qq->xmflg[i] = 0;
	    memcpy(&qq->last[0].xmsg, qq->xmbuf+i, sizeof(XMSG));
	    return qq->xmbuf+i;
	}

    /* Find the least recently used buffer and use that */

    i = 0;
    for( j = 0; j < MAX_XMSG; ++j )
	if( qq->xmnum[j] < qq->xmnum[i] )
	    i = j;
    if( qq->xmflg[i] )
    {
	Verbose(4,"QQLINK","readxmsg: flushing message %ld (%d)",qq->xmnum[i],i);
	flushxmsg(qq, i);
    }
    MsgReadMsg(hmsg, qq->xmbuf+i, 0, 0, NULL, 0, NULL);
    qq->xmnum[i] = msgn;
    qq->xmflg[i] = 0;
    memcpy(&qq->last[0].xmsg, qq->xmbuf+i, sizeof(XMSG));
    return qq->xmbuf+i;
}




/*# ----------------------------------------------------------------------
 * writexmsg(qq, hmsg, msgn, xmsg)
 *
 * PARAMETER
 *	qq		data of related .QQL file
 *	hmsg		handle of opened message 'msgn'
 *			(or NULL:  writexmsg() will open the message)
 *	msgn		message no to write
 *	xmsg		XMSG data to write
 *
 * RETURNS
 *	(nothing)
 *
 * DESCRIPTION
 * REMARKS
 */
PRIVATE void
writexmsg(PQQDAT const qq, HMSG hmsg, long const msgn, PXMSG const xmsg)
{
    int i;

    /* Check for a redundant write */

    if( msgn == qq->last[0].xnum )
    {
	if( 0 == memcmp(xmsg, &qq->last[0].xmsg, sizeof(XMSG)) )
	    return;
	qq->last[0].xnum = 0;
    }

    /* Write the xmsg to buffer or disk */

    for( i = 0; i < MAX_XMSG; ++i )
	if( qq->xmnum[i] == msgn )
	    break;
    if( i >= MAX_XMSG  ||  hmsg != NULL )
    {
	int	i_opened = 0;

	Verbose(4,"QQLINK","writexmsg: writing msg no %ld", msgn);
	if( hmsg == NULL )
	{
	    /*printf("Internal error in writexmsg()\n");*/
	    hmsg = MsgOpenMsg(qq->area, MOPEN_RW, msgn);
	    ++i_opened;
	}
	if( hmsg != NULL )
	{
	    MsgWriteMsg(hmsg, 0, xmsg, NULL, 0, 0, 0, NULL);
	    if( i < MAX_XMSG )
		qq->xmflg[i] = 0;
	    if( i_opened )
		MsgCloseMsg(hmsg);
	}
    }
    else
	qq->xmflg[i] = 1;
    return;
}





/*# ----------------------------------------------------------------------
 * getsubj(qq, hmsg, msgn, *to_all)
 *
 * PARAMETER
 *	*qq		data of related .QQL file
 *	hmsg		handle of open message 'msgn'
 *	msgn		msgn of message to build MSGID from
 *	*to_all
 *
 * RETURNS
 *	QQLink's unique MSGID
 *
 * DESCRIPTION
 *
 * REMARKS
 */
PRIVATE long
getsubj(PQQDAT qq, HMSG hmsg, long msgn, int *to_all)
{
    PXMSG	xmsg = readxmsg(qq, hmsg, msgn);
    char *		s = xmsg->subj;
    long		id = 0;
    extern char *	all_name;

    /* Strip 'RE:'s and leading spaces. */

    while( *s )
    {
	if( !strnicmp(s,"re:", 3) )
	    s += 3;
	else if( *s == ' ' )
	    ++s;
	else
	    break;
    }

    /* Encode the subj. */

    makeid(&id, s);
    if( all_name  &&  *all_name )
	*to_all = !stricmp(xmsg->to, all_name);
    else
	*to_all = 0;
    return id;
}




/*# ----------------------------------------------------------------------
 * setorig(*qq, *msg, msgn, uid)
 *
 * PARAMETER
 *	*qq		data of related .QQL file
 *	*msg		handle of opened message 'msgn'
 *			(which is reply to 'uid')
 *	msgn		message no of reply message
 *	uid		UID of original message
 *
 * RETURNS
 *	(nothing)
 *
 * DESCRIPTION
 *	Reads message '*msg'/'msgn' into memory, adds 'replyto = uid'
 *	field to it and writes the message back.
 *
 * REMARKS
 */
PRIVATE void
setorig(PQQDAT qq, HMSG hmsg, long msgn, long uid)
{
    int		i_opened_msg = 0;
    PXMSG	xmsg;

    xmsg = readxmsg(qq, hmsg, msgn);
    if( xmsg == NULL  &&  hmsg == NULL )
    {
	hmsg = MsgOpenMsg(qq->area, MOPEN_RW, msgn);
	++i_opened_msg;
	xmsg = readxmsg(qq, hmsg, msgn);
    }
    if( xmsg == NULL )
	return;					/* still a problem? */

    xmsg->replyto = uid;
    writexmsg(qq, hmsg, msgn, xmsg);
    if( i_opened_msg )
	MsgCloseMsg(hmsg);
    return;
}




/*# ----------------------------------------------------------------------
 * setrepl(*qq, *msg, msgn, uid)
 *
 * PARAMETER
 *	*qq		data of related .QQL file
 *	hmsg		handle of opened message 'msgn' or NULL
 *	msgn		message no of original message
 *	uid		UID of original message
 *
 * RETURNS
 * DESCRIPTION
 * REMARKS
 */
PRIVATE void
setrepl(PQQDAT qq, HMSG hmsg, long msgn, long uid)
{
    int 	i;
    int		i_opened_msg = 0;
    PXMSG	xmsg;

    xmsg = readxmsg(qq, hmsg, msgn);
    if( xmsg == NULL  &&  hmsg == NULL )
    {
	hmsg = MsgOpenMsg(qq->area, MOPEN_RW, msgn);
	++i_opened_msg;
	xmsg = readxmsg(qq, hmsg, msgn);
    }
    if( xmsg == NULL )
	return;					/* still a problem? */

    for( i = 0; i < MAX_REPLY; ++i)
	if( xmsg->replies[i] == 0  ||  xmsg->replies[i] == uid )
	    break;
    i %= MAX_REPLY;
    xmsg->replies[i] = uid;
    writexmsg(qq, hmsg, msgn, xmsg);
    if( i_opened_msg )
	MsgCloseMsg(hmsg);
    return;
}




#define ADJ(x)	(((x)+MAX_MSGS)%MAX_MSGS)

/*# ----------------------------------------------------------------------
 * upd_table(*qq, id, uid)
 *
 * PARAMETER
 *	*qq
 *	id		QQLINK's unique code of MSGID kludge
 *	orig		QQLINK's unique code of REPLY kludge
 * 	uid		squish's unique message id
 *
 * RETURNS
 *	'id' as passed
 *
 * DESCRIPTION
 *	Adds 'id' and 'uid' to end of tables in 'qq'.
 *
 * REMARKS
 */
PRIVATE MSGID
upd_table(struct QQF * const qq, MSGID const own, MSGID const orig, long const uid)
{
    qq->own_qqid[qq->topid] = own;
    qq->orig_qqid[qq->topid] = orig;
    qq->uid[qq->topid] = uid;
    qq->high = uid;
    qq->topid = ADJ(qq->topid+1);
    return own;
}






/* **********************************************************************
 * **** Public Entries **************************************************
 * ******************************************************************* */


PUBLIC PQQDAT
open_qq(const char *path, HAREA const area, int is_squish, int method)
{
    char	buf[80];
    PQQDAT	qq;

    /* Make a path to the QQLink info file for this area */

    strcpy(buf, path);
    if( is_squish )
	strcat(buf, ".QQL");
    else
	strcat(buf, "\\QQLINK.");

    /* Set up qq area struct */

    if( area == NULL )
	return NULL;
    qq = calloc(1, sizeof(QQDAT));
    if( qq == NULL )
	return qq;
    qq->area = area;
    if( !link_slow )
    {
	qq->file = _open(buf, O_CREAT|O_BINARY|O_RDWR, S_IREAD|S_IWRITE);
	if( (short)qq->file < 0 )
	    return NULL;
	_read(qq->file, &qq->f, sizeof(struct QQF));

	/* If combined mode, (try to) load extra table */
	if( qq->f.method == LINK_COM1 )
	    _write(qq->file, &qq->f2, sizeof(struct QQF));
    }

    /* Old version, or with wrong method, or doesn't exist? start anew! */

    if( qq->f.sig != QQL2_SIG
	||  (method != qq->f.method  &&  method != LINK_LAST)
	||  link_slow )
    {
	memset(&qq->f, 0, sizeof(struct QQF));
	memset(&qq->f2, 0, sizeof(struct QQF));
	qq->f.sig = QQL2_SIG;
	qq->f.method = LINK_FTS9;
	qq->f2.sig = QQL2_SIG;
	qq->f2.method = LINK_COM2;
    }
    if( method != LINK_LAST )
	qq->f.method = method;
    qq->oldhigh = qq->f.high;
    return qq;
}




/*# ----------------------------------------------------------------------
 * link_qq(*_qq, *msg, uid, clean)
 *
 * PARAMETER
 *	*_qq		data of related .QQL file
 *	hmsg		handle of opened message
 *			(the one which is to be linked)
 *	uid		'unique message ID' of message to be linked
 *	clean		/0 = remove all old links first
 *
 * RETURNS
 *	0		not linked (no answer/reply found)
 *	>0		messages linked (more than one if 'uid' is source
 *			of several replies)
 *
 *	-2		no REPLY or MSGID kludge
 *
 * DESCRIPTION
 * REMARKS
 *	We don't analyse a REPLY kludge if there isn't a MSGID. But I don't
 *	think this situation even exists.
 */
PUBLIC int
link_qq(PQQDAT const _qq, HMSG const hmsg, long const uid, int const clean)
{
    int		result = 0;			/* default: no matching message */
    MSGID	orig = 0, msgid = 0, auxid = 0;

    /* Set up a couple of shortcuts: */

    HAREA const	area = _qq->area;
    struct QQF *qq = &(_qq->f);
    long const	msgn = MsgUidToMsgn(area, uid, UID_EXACT);
    int		to_all;


    /* if we're cleaning, wipe this message's reply links. */

    if( clean )
    {
	PXMSG const	xmsg = readxmsg(_qq, hmsg, msgn);
	memset(xmsg->replies, 0, sizeof(long)*MAX_REPLY );
	xmsg->replyto = 0;
	writexmsg(_qq, hmsg, msgn, xmsg);
    }


    /* Get ID and ORIGID (for subject linking those are identical) */

    switch( qq->method )
    {
      case LINK_FTS9:
	getid(hmsg, &msgid, &orig);
	if( upd_table(qq, msgid, orig, uid) == 0 )
	    return -2;				/* No MSGID */
	break;

      case LINK_SUBJ:
	orig = msgid = getsubj(_qq, hmsg, msgn, &to_all);
	if( to_all )
	    orig = 0;
	upd_table(qq, msgid, 0, uid);
	break;

      case LINK_COM1:
	getid(hmsg, &msgid, &orig);
	auxid = getsubj(_qq, hmsg, msgn, &to_all);
	upd_table(qq, msgid, orig, uid);
	upd_table(&_qq->f2, auxid, 0, uid);
	if( !orig  &&  !to_all )
	{
	    msgid = orig = auxid;
	    qq = &_qq->f2;
	}
	break;
    }


    /* If there's an original (a REPLY kludge inside 'hmsg'),
     * try to link to that one.  The MSGID of the original is
     * contained in 'orig'. */

    if( orig )
    {
	int	i;

	/* Scan backwards from the current message for the message which
	 * is referenced in our REPLY kludge ('orig'). */

	for( i = ADJ(qq->topid-2); i != qq->topid; i = ADJ(i-1) )
	    if( qq->own_qqid[i] == orig )
		break;

	/* Found one? */

	if( i != qq->topid )
	{
	    long const	omsgn = MsgUidToMsgn(area, qq->uid[i], UID_EXACT);

	    if( omsgn < 0  ||  omsgn == msgn )
		result = 2;
	    else
	    {
		setorig(_qq, hmsg, msgn, qq->uid[i]);
		setrepl(_qq,/*omsg*/NULL, omsgn, uid);
		++result;			/* found original */
	    }
	}
    }


    /* If our current message has an ID, search for others to reference
     * it via their REPLY kludge.  This is for situations when the answer
     * is received before the question arrives. */

    if( msgid )
    {
	int	i;

	/* Scan backwards from the current message for all messages
	 * where 'msgid' is referenced in the REPLY kludge. */

	for( i = ADJ(qq->topid-2); i != qq->topid; i = ADJ(i-1) )
	    if( qq->orig_qqid[i] == msgid )
	    {
		/* Found one! */

		long const	rmsgn = MsgUidToMsgn(area, qq->uid[i], UID_EXACT);

		if( !(rmsgn < 0  ||  rmsgn == msgn) )
		{
		    setorig(_qq,/*omsg*/NULL, rmsgn, uid);
		    setrepl(_qq, hmsg, msgn, qq->uid[i]);
		    ++result;			/* found reply */
		}
	    }
    }

    return result;
}




PUBLIC void
close_qq(PQQDAT qq)
{
    int	i, c = 0;

    if( qq == NULL )
	return;
    for( i = 0; i < MAX_XMSG; ++i )
	c += flushxmsg(qq, i);
    Verbose(4,"QQLINK","close_qq: flushed %d messages", c);
    if( !link_slow )
    {
	if( qq->f.high != qq->oldhigh )
	{
	    lseek(qq->file, 0, 0);
	    _write(qq->file, &qq->f, sizeof(struct QQF));

	    /* if combined mode, save extra table */

	    if( qq->f.method == LINK_COM1 )
		_write(qq->file, &qq->f2, sizeof(struct QQF));
	    _write(qq->file, NULL, 0);	/* make sure we truncate file here. */
	}
	_close(qq->file);
    }
    free(qq);
}




PUBLIC int
percent_qq(PQQDAT qq, long msgnum)
{
    int i, nided = 0;

    if( msgnum <= 0 )
	return 0;
    if( msgnum > MAX_MSGS )
	msgnum = MAX_MSGS;
    for( i = 0; i < msgnum; ++i )
	if( qq->f.own_qqid[ADJ(qq->f.topid-i)] )
	    ++nided;
    return ((long)nided * 100) / msgnum;
}
