/*
 * $Source: r:/source/msgapi/qqlink/RCS/qqdat.h,v $
 * $Revision: 2.3 $
 * $Date: 2000/11/21 02:09:01 $
 * $Author: vitus $
 *
 * QQLink 1.24.jx - .QQL file handling
 *
 * $Log: qqdat.h,v $
 * Revision 2.3  2000/11/21 02:09:01  vitus
 * - changed MSGH* to HMSG and MSG * to HAREA (as sqdev suggests)
 *
 * Revision 2.2  2000/11/17 02:21:27  vitus
 * - moved last_* (detection of unneeded writes) to QQDAT structure.
 *
 * Revision 2.1  2000/11/16 02:59:24  vitus
 * - struct QQF now contains IDs of MSGID and REPLY
 *
 * Revision 1.3  1998/09/05 18:13:41  vitus
 * - added type PMSGID
 *
 * Revision 1.2  1997/09/16 10:26:49  vitus
 * - beautified
 */

/**
***	This file is part of the source to QQLink, a FTS-9 reply linker for
*** use with the Squish(tm) mail processor.
*** Copyright (c) 1993 Josh Parsons. All rights reserved, however:
***
***	       1) You are permitted to freely distribute this source,
***           provided you do not remove this copyright notice, and
***           provided you do not charge a fee, for the service of
***           distributing this material, over and above the value
***           of disks, phone tolls, etc. incurred in such distribution.
***        2) You are permitted to use executables produced from this
***           source in the manner described in QQLink.DOC, which should
***           accompany this file.
***        3) You are permitted to use this source in your own programs,
***           provided that your derivative work carries an
***           acknowledgement that portions of its source are derived
***           from QQLink, by Josh Parsons.
***
***        NB: This source carries no warrantee, implicit or explicit
***            of fitness for any particular task.
**/


#define MAX_XMSG	192
/*#define MAX_XMSG	32*/
#define MAX_MSGS	1000			/* 16 bit! */

typedef long		MSGID;
typedef MSGID *		PMSGID;



/* The actual layout of a .QQL file */
struct QQF {
    byte	sig;
    byte	method;
    word	topid;
    ulong	high;
    MSGID	own_qqid[MAX_MSGS];		/* MSGID */
    MSGID	orig_qqid[MAX_MSGS];		/* REPLY */
    ulong	uid[MAX_MSGS];
};



/* One muthahumper of a structure (62682 bytes) */
typedef struct _QQDAT {
    word	file;
    HAREA	area;
    ulong	oldhigh;

    /* The following three elements build a read cache of
     * the most recently used messages, only XMSG. */

    XMSG	xmbuf[MAX_XMSG];		/* administrative data of messages */
    ulong	xmnum[MAX_XMSG];		/* message no of message */
    byte	xmflg[MAX_XMSG];		/* /0 = the entry is dirty */


    /* 'last': used to suppress redundant writes.
     *
     * It will get a copy of the last read XMSG and its msgn.  When
     * it's time to write we will check if the new XMSG is still the
     * same as the one contained in 'last'.  If it's the same, we do
     * not write. */

    struct {
	XMSG	xmsg;
	ulong	xnum;
    } last[1];					/* more than one? */


    struct QQF	f;
    struct QQF	f2;
} QQDAT, * PQQDAT;



#define LINK_LAST	0
#define LINK_FTS9	1
#define LINK_SUBJ	2
#define LINK_COM1	3
#define LINK_COM2	4

#define QQL2_SIG	('Q'+1)

extern int	trunc_qq;


extern PQQDAT	open_qq(const char *path, HAREA area, int is_squish, int method);
extern int	link_qq(PQQDAT qq, HMSG hmsg, long uid, int clean);
extern void	close_qq(PQQDAT qq);
extern int	percent_qq(PQQDAT qq, long msgnum);
