/*
 * $Source: r:/source/msgapi/qqlink/RCS/qqlink.c,v $
 * $Revision: 1.19 $
 * $Date: 2001/06/12 00:04:40 $
 * $Locker:  $
 *
 *	QQLink 1.24.jx - main module
 *
 * $Log: qqlink.c,v $
 * Revision 1.19  2001/06/12 00:04:40  vitus
 * - link_area: changed code back to rev1.16, 1.17 and 1.18 destroyed some
 *   indices on my testpoint
 *
 * Revision 1.18  2001/06/06 00:12:45  vitus
 * - v1.24.j9 (don't create or truncate area files)
 * - "-?" lists all commandline options (well, all starting with "-")
 *
 * Revision 1.17  2001/05/30 00:54:26  vitus
 * - link_area: open area in mode MSGAREA_NORMAL not MSGAREA_CRIFNEC because
 *   this will destroy an area in error!
 * - link_area: more return value checks
 *
 * Revision 1.16  2000/12/12 01:53:23  vitus
 * - v1.24.j8 (gecho structure packing)
 *
 * Revision 1.15  2000/11/21 02:11:05  vitus
 * - changed MSGH* to HMSG and MSG * to HAREA (as sqdev suggests)
 * - modifications because of changed result of link_qq()
 *
 * Revision 1.14  2000/11/16 03:03:28  vitus
 * - minor change: QQDAT * to PQQDAT
 * - V1.24.j7 (reverse linking)
 *
 * Revision 1.13  2000/05/29 02:31:56  vitus
 * - changed version no to 1.24.j6
 *
 * Revision 1.12  1999/07/01 23:50:22  vitus
 * - read_echolog/read_antilog: read area tags into MAX_TAGLEN buffer
 * - includes qqlink.h, local declarations removed
 *
 * Revision 1.11  1998/12/19 00:44:58  vitus
 * - changed directory of common header files
 *
 * Revision 1.10  1998/11/23 00:33:24  vitus
 * - checks *any* call to OS/2 or MSGAPI
 * - commandline options for Maximus and GEcho configuration files
 *
 * Revision 1.9  1998/09/22 23:49:37  vitus
 * - waits until all threads ended bwfore closing msgapi
 * - handles "thread not started" condition
 *
 * Revision 1.8  1998/09/05 17:04:29  vitus
 * - changed version no to 1.24.j3
 * - added option -P<n>, no of threads
 * - updated help screen
 *
 * Revision 1.7  1998/07/28 20:36:06  vitus
 * - corrected statistics (no to link)
 *
 * Revision 1.6  1998/06/05 09:38:24  vitus
 * - removed all colours in screen output
 * - switched from printf() to Verbose()
 * - modified -q[q] option to -q[n]
 * - added -b[n]file to specify binkley-style logfile
 *
 * Revision 1.5  1998/01/31 02:41:40  vitus
 * - added multithreading (still quick hack)
 *
 * Revision 1.4  1998/01/31 00:07:55  vitus
 * - added -q and -qq options
 * - cleanup of help()
 * - removed some unused code
 *
 * Revision 1.3  1997/09/17 01:19:34  vitus
 * marked all symbols either PRIVATE or PUBLIC
 * reset attributes before calling help()
 *
 * Revision 1.2  1997/09/16 10:22:31  vitus
 * beautified
 */
static char const id[]="$Id: qqlink.c,v 1.19 2001/06/12 00:04:40 vitus Exp $";

/**
***  This file is part of the source to QQLink, a FTS-9 reply linker for
*** use with the Squish(tm) mail processor.
*** Copyright (c) 1993 Josh Parsons. All rights reserved, however:
***
***         1) You are permitted to freely distribute this source,
***           provided you do not remove this copyright notice, and
***           provided you do not charge a fee, for the service of
***           distributing this material, over and above the value
***           of disks, phone tolls, etc. incurred in such distribution.
***        2) You are permitted to use executables produced from this
***           source in the manner described in QQLink.DOC, which should
***           accompany this file.
***        3) You are permitted to use this source in your own programs,
***           provided that your derivative work carries an
***           acknowledgement that portions of its source are derived
***           from QQLink, by Josh Parsons.
***
***        NB: This source carries no warrantee, implicit or explicit
***            of fitness for any particular task.
**/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#if defined(__MT__) || defined(__MULTI__)
# define INCL_DOS
# define INCL_ERRORS
# include <os2.h>
#endif

#include "compiler.h"				/* let MSGAPI analysze compiler */
#include "MsgApi.h"

#include "../../Lib/defines.h"			/* my standard defines */
#include "../../Lib/verbose.h"

#include "qqlink.h"
#include "qqdat.h"


/* Definitions to modify any logo QQLink writes to screen. */

#define VERSION_STR	"V1.24.j9"
/*#define PORTED_BY	"Your name here"*/
/*#define PORTED_TO	"OS/2"*/
#define MODIFIED_BY	"Vitus Jensen (2:2474/424)"




#if defined(__MT__) || defined(__MULTI__)
/* This range compiles if multithreaded operation is enabled. */

# define NUM_THREADS		10
# define QUEREQ_PROCESS		0
# define QUEREQ_EXIT		1

#define SZ_QUEAREA		"/queues/qqlink/area"
#define SZ_QUECOUNT		"/queues/qqlink/msgcount"

PRIVATE HQUEUE	hqArea;
PRIVATE HQUEUE	hqCount;
PRIVATE PID	pidAreaQueue;
PRIVATE PID	pidCountQueue;
#endif




typedef struct area_t {
    struct area_t *next;			/* linked list */
    char *	tag;				/* echotag */
    char *	path;				/* path */
    char	is_sq;				/* area is .SQ format */
    char	dolink;				/* area is tagged to be linked */
    char	subj;				/* future expansion */
    char	is_net;				/* is a netmail area */
} AREA;






/* Symbols exported from this module. */

PUBLIC char	link_passthru = 0;
PUBLIC char	link_slow = 0;

PUBLIC char *	all_name = "";




/* Local symbols, not exported. */

PRIVATE char	kill_log = 0;
PRIVATE char	kill_qql = 0;
PRIVATE char	link_all = 0;
PRIVATE char	link_meth = LINK_FTS9;
PRIVATE char	cfg_read = 0;
PRIVATE char	link_clean = 0;
PRIVATE char	link_prune = 0;
PRIVATE UINT	cThreads = NUM_THREADS;

PRIVATE char	* outlog_fname = NULL;

PRIVATE AREA	* rootarea = NULL;




PUBLIC void
add_area(const char *tag,const char *path,int is_sq,int is_net)
{
    AREA *ar = calloc(1, sizeof(AREA));
    if( ar == NULL )
	return;
    ar->tag = strdup(tag);
    ar->path = strupr(strdup(path));
    if( ar->path[strlen(ar->path) - 1] == '\\' )
	ar->path[strlen(ar->path) - 1] = 0;
    ar->is_sq = (char)is_sq;
    ar->is_net = (char)is_net;
    if( link_all )
	ar->dolink = 1;
    else
	ar->dolink = 0;
    ar->next = rootarea;
    rootarea = ar;
}




PRIVATE AREA *
findarea(const char *tag)
{
    AREA *ar = rootarea;

    for(; ar ; ar = ar->next )
	if( !stricmp(ar->tag, tag) )
	    break;
    return ar;
}




/**
*** To do: message expiry here...
**/

PRIVATE int
do_prune(HAREA const area, int const is_sq)
{
#if 0
    HMSG hmsg = NULL;
    long i, maxmsgs = 0, minmsgs = 0, maxdays = 0;

    if( is_sq )
	SquishGetMaxMsg(area, (dword *)&maxmsgs, (dword *)&minmsgs,
			(dword *)&maxdays);
    if( MsgHighMsg(area) < minmsgs )
	return 0;
    if( maxdays < 1 )
	return 0;
    for( i = 1; i < MsgHighMsg(area); ++i )
    {
	msg = MsgOpenMsg(area, MOPEN_RW, i);
	if( msg )
	    break;
    }
    if( msg == NULL )
	return 0;
#endif
    return 0;
}




#define CENT(num,den)	(int)(((long)num*100)/den)

PRIVATE void
link_upd(char * tag, long msgn, long msgh, int nmsg, int num2link)
{
    int	cent;
    if( num2link == 0 )
	nmsg = num2link = 1;			/* Avoid dividing by zero */
    cent = CENT(nmsg, num2link);
    if( cent > 100 )
	cent = 100;
    if( cent < 0 )
	cent = 0;
    Verbose(4,"QQLINK", "Area %-40s, done %04ld/%04ld (%02d%%)",tag,msgn,msgh,cent);
    return;
}




/*# ----------------------------------------------------------------------
 * link_area(*tag, *name, is_sq, is_net)
 *
 * PARAMETER
 *	*tag		area tag
 *	*name		
 *	is_sq		true = area in squish format
 *	is_net		true = netmail area
 *
 * RETURNS
 *	count of messages linked
 *
 * DESCRIPTION
 * REMARKS
 */
PRIVATE int
link_area(char *tag, char *name, int is_sq, int is_net)
{
    HAREA	area;
    PQQDAT	qq = NULL;
    int		nlinked = 0, nmsg = 0;
    int		num2link;
    long	msgn, msgh;
    time_t	tt = 0;
    sword	error = 0;

    Verbose(3, "QQLINK", "Opening %-40s", tag);

    area = MsgOpenArea(name, MSGAREA_NORMAL,
		       (is_sq ? MSGTYPE_SQUISH : MSGTYPE_SDM)
		       | ((is_net || is_sq) ? 0 : MSGTYPE_ECHO));
    if( area )
	qq = open_qq(name, area, is_sq, link_meth);
    if( area )
	error = MsgLock(area);
    if( area == NULL  ||  qq == NULL  || error != 0 )
    {
	Verbose(0, "QQLINK", "Error opening %-40s", name);
	if( qq != NULL )
	    close_qq(qq);
	if( area )
	{
	    MsgUnlock(area);
	    MsgCloseArea(area);
	}
	return 0;
    }

    Verbose(3, "QQLINK", "Reading %-40s", tag);

    /* OK, 'area' and 'qq' are opened, locked and valid.
     * Do real linking. */

    msgh = MsgGetHighMsg(area);
    if( kill_qql )
    {
	qq->oldhigh = 0;
	msgn = 0;
    }
    else
	msgn = MsgUidToMsgn(area, qq->f.high, UID_NEXT);

    num2link = msgh - msgn + 1;
    link_upd(tag, msgn, msgh, nmsg, num2link);
    while( msgn <= msgh )
    {
	HMSG hmsg = MsgOpenMsg(area, MOPEN_RW, msgn);
	if( hmsg )
	{
	    int	i = link_qq(qq, hmsg, MsgMsgnToUid(area, msgn), link_clean);
	    ++nmsg;
	    if( i > 0 )
		nlinked += i;
	    else switch( i )
	    {
	      case -2:				/* message does not use FTS-9 */
		break;
	    }

	    MsgCloseMsg(hmsg);
	}
	if( time(NULL) > tt )
	{
	    link_upd(tag, msgn, msgh, nmsg, num2link);
	    time(&tt);
	}
	++msgn;
    }
    link_upd(tag, msgn-1, msgh, nmsg, num2link);


    {
	int	id_pc = percent_qq(qq, msgh);

	Verbose(3, "QQLINK", "Writing %-40s", tag);

	close_qq(qq);
	if( is_sq  &&  link_prune )
	{
	    long	n_pruned = 0;
	    long const	n2prune = MsgHighMsg(area);

	    tt = 0;
	    while( do_prune(area, is_sq) )
	    {
		++n_pruned;
		if( time(NULL) > tt )
		{
		    time(&tt);
		    Verbose(4, "Area %-40s  %04ld/%04ld (pruning) ",
			    tag, n_pruned, n2prune);
		}
	    }
	}
	if( link_meth == LINK_SUBJ )
	    Verbose(1, "QQLINK", "Area %-40s  %04d/%04d linked",
		    tag, nlinked, num2link);
	else
	    Verbose(1, "QQLINK", "Area %-40s  %04d/%04d linked  %3d%% MSGID",
		    tag, nlinked, num2link, id_pc);
    }


    /* Close 'area', 'qq' was already closed. */

    MsgUnlock(area);
    MsgCloseArea(area);

    return nmsg;
}




PRIVATE int
link_echo(AREA *ar)
{
    int	n;

    Verbose(3, "QQLINK", "Linking %-40s", ar->tag);
    n = link_area(ar->tag, ar->path, ar->is_sq, ar->is_net);
    return n;
}




#if defined(__MT__) || defined(__MULTI__)
/*#
 * CALL
 *	LinkerThread(dummy)
 * PARAMETER
 *	dummy		ignored
 * RETURNS
 *	(nothing)
 * GLOBAL
 *	pidAreaQueue, hqArea, hqCount
 * DESCRIPTION
 * REMARKS
 */
PRIVATE VOID
LinkerThread(PVOID dummy)
{
    APIRET	rc;
    REQUESTDATA	reqdta;
    ULONG	dtalen;
    BYTE	prio;
    AREA *	ar;
    ULONG	count;

    Verbose(4, "QQLINK", "LinkerThread started");
    for( rc = 0; rc == 0 ;)
    {
	reqdta.pid = pidAreaQueue;
	rc = DosReadQueue(hqArea, &reqdta,&dtalen,(PPVOID)&ar, 0, DCWW_WAIT,&prio,0);
	if( rc == 0 )
	    switch( reqdta.ulData )
	    {
		/* QUEREQ_PROCESS:
		 *	Link area.
		 *	Write count of linked messages back to
		 *	the 'hqCount' queue. */
	      case QUEREQ_PROCESS:
		count = link_echo(ar);
		do
		{
		    rc = DosWriteQueue(hqCount, count, sizeof(*ar), ar, 0);
		    if( rc == ERROR_QUE_NO_MEMORY )
		    {
			Verbose(3, "QQLINK", "Queue busy, wait and retry");
			DosSleep(100);		/* wait and retry */
		    }
		    else if( rc != 0 )
		    {
			Verbose(0, "QQLINK", "DosWriteQueue - rc %lu", rc);
			break;
		    }
		}
		while( rc != 0 );
		break;

		/* QUEREQ_EXIT:
		 *	End loop.
		 *	Write something to the 'hqCount' queue so
		 *	the main thread can detect when all threads
		 *	have stopped using the msgapi (in case it
		 *	would be a problem calling 'MsgCloseApi'. */
	      case QUEREQ_EXIT:
		do
		{
		    rc = DosWriteQueue(hqCount, 0, 0, (PVOID)0xFFFFFFFF, 0);
		    if( rc == ERROR_QUE_NO_MEMORY )
		    {
			Verbose(3, "QQLINK", "Queue busy, wait and retry");
			DosSleep(100);
		    }
		    else if( rc != 0 )
		    {
			Verbose(0, "QQLINK", "DosWriteQueue - rc %lu", rc);
			break;
		    }
		}
		while( rc != 0 );
		rc = 0xFFFF;			/* any value != 0 */
		break;
	    }
    }

    Verbose(4, "QQLINK", "LinkerThread ended");
    return;
}




PRIVATE APIRET
DoLinking(UINT *alinked,PULONG const nlinked)
{
    AREA *	ar;
    REQUESTDATA	reqdta;
    ULONG	dtalen;
    BYTE	prio;
    UINT	i;
    APIRET	rc;


    /* Depending on '-P<n>' parameter we may need more than the default
     * 50 file handles OS/2 gives us as a default (may be already incremented).
     * Solution: just try to set the limit.  If it's less than the current
     * value OS/2 ignores the request, if it's more OS/2 expands it's tables.
     * Assume 5 files per area/linking thread plus some overhead. */

    rc = DosSetMaxFH(10 + cThreads * 9);
    if( rc != 0  &&  rc != ERROR_INVALID_PARAMETER )
    {
	LONG	l = 0;
	ULONG	current = 0;

	rc = DosSetRelMaxFH(&l, &current);
	if( rc != 0 )
	{
	    Verbose(0, "QQLINK", "DosSetRelMaxFH - rc %lu", rc);
	    return 3;
	}
	cThreads = (current - 10) / 5;
	Verbose(0, "QQLINK",
		"Couldn't increase handle count, restricted to %u threads",cThreads);
    }


    /* Create two queues used to communicate with LinkerThreads.  For
     * some reasons all samples pass the PID of the queue's owner to
     * DosReadQueue() and therefore call DosOpenQueue() after creation;
     * I don't really know why but do the same here. */

    rc = DosCreateQueue(&hqArea, QUE_FIFO|QUE_NOCONVERT_ADDRESS, SZ_QUEAREA);
    if( rc == 0 )
	rc = DosCreateQueue(&hqCount, QUE_FIFO|QUE_NOCONVERT_ADDRESS, SZ_QUECOUNT);
    if( rc )
    {
	Verbose(0, "QQLINK", "DosCreateQueue - rc %lu", rc);
	return 3;
    }
    rc = DosOpenQueue(&pidAreaQueue, &hqArea, SZ_QUEAREA);
    if( rc == 0 )
	rc = DosOpenQueue(&pidCountQueue, &hqCount, SZ_QUECOUNT);
    if( rc )
    {
	Verbose(0, "QQLINK", "DosOpenQueue - rc %lu", rc);
	return 3;
    }


    /* Start worker threads */

    Verbose(3, "QQLINK", "Starting linker threads");
    for( i = cThreads; i > 0; --i )
    {
	int tid = _beginthread(LinkerThread, NULL, 128*1024, NULL);

	if( tid == -1 )
	{
	    Verbose(0, "QQLINK", "Can't start another thread, ignored");
	    --cThreads;
	}
    }


    /* For every area to be linked, write a command to
     * our area queue.  Worker threads will read from queue and
     * do the linking. */

    Verbose(3, "QQLINK", "Passing area infos to threads");
    for( ar = rootarea; ar; ar = ar->next )
	if( ar->dolink )
	{
	    do
	    {
		rc = DosWriteQueue(hqArea, QUEREQ_PROCESS, sizeof(*ar), ar, 0);
		if( rc == ERROR_QUE_NO_MEMORY )
		{
		    Verbose(3, "QQLINK", "Queue busy, wait and retry");
		    DosSleep(100);
		}
		else if( rc != 0 )
		{
		    Verbose(0, "QQLINK", "DosWriteQueue - rc %lu", rc);
		    abort();
		}
	    }
	    while( rc != 0 );
	    ++(*alinked);
	}


    /* Read result codes.  There is a code for every posted area. */

    Verbose(3, "QQLINK", "Reading counts for %u areas", *alinked);
    for( i = 0; i < *alinked; ++i )
    {
	ULONG	count;
	AREA *	area;

	reqdta.pid = pidCountQueue;
	rc = DosReadQueue(hqCount, &reqdta, &dtalen, (PPVOID)&area,
			  0, DCWW_WAIT, &prio, 0);
	if( rc != 0 )
	{
	    Verbose(0, "QQLINK", "DosReadQueue - rc %lu", rc);
	    abort();
	}
	count = reqdta.ulData;
	Verbose(4, "QQLINK", "A thread reported '%s, %lu linked'", area->tag, count);
	*nlinked += count;
    }


    /* Stop worker threads */

    Verbose(3, "QQLINK", "Passing EXIT command to all threads");
    for( i = 0; i < cThreads; ++i )
	do
	{
	    rc = DosWriteQueue(hqArea, QUEREQ_EXIT, 0, NULL, 0);
	    if( rc == ERROR_QUE_NO_MEMORY )
	    {
		Verbose(3, "QQLINK", "Queue busy, wait and retry");
		DosSleep(100);
	    }
	    else if( rc != 0 )
	    {
		Verbose(0, "QQLINK", "DosWriteQueue - rc %lu", rc);
		break;
	    }
	}
	while( rc != 0 );


    /* Wait for all threads having completed their linking. */

    Verbose(3, "QQLINK", "Waiting for a 'complete' from all threads");
    for( i = 0; i < cThreads; ++i )
    {
	ULONG	count;

	rc = DosReadQueue(hqCount, &reqdta, &dtalen,
			  (PVOID)&count, 0, DCWW_WAIT, &prio, 0);
	if( rc != 0 )
	{
	    Verbose(0, "QQLINK", "DosReadQueue - rc %lu", rc);
	    abort();
	}
	Verbose(4, "QQLINK", "A thread reported 'ended'");
    }

    Verbose(3, "QQLINK", "All threads completed, done linking");
    return 0;
}




#else  /* ?__MT/MULTI__ */

PRIVATE APIRET
DoLinking(UINT *alinked,PULONG const nlinked)
{
    for( ar = rootarea; ar; ar = ar->next )
	if( ar->dolink )
	{
	    *nlinked += link_echo(ar);
	    ++(*alinked);
	}
    return 0;
}
#endif /* ?__MT/MULTI__ */




#define PARJCMP(x,y)	strncmp(x,y,strlen(y))
#define PARCMP(x,y)	strnicmp(x,y,strlen(y))
#define PARGET(x,y)	((x)+strlen(y))

PRIVATE void
read_echolog(const char *echolog_fname,int kill)
{
    FILE *	fin;
    char	buf[MAX_TAGLEN];

    if( echolog_fname == NULL )
	return;
    if( !cfg_read )
    {
	q_config();
	++cfg_read;
    }
    fin = fopen(echolog_fname, "r");
    if( fin == NULL )
	return;
    while( !feof(fin) )
    {
	AREA *	ar;

	fgets(buf, MAX_TAGLEN, fin);
	if( strchr(buf, '\n') )
	    *strchr(buf, '\n') = 0;
	ar = findarea(buf);
	if( ar )
	    ar->dolink = 1;
    }
    fclose(fin);
    if( kill )
	remove(echolog_fname);
}




PRIVATE void
read_antilog(const char *echolog_fname,int kill)
{
    FILE *	fin;
    char	buf[MAX_TAGLEN];

    if( echolog_fname == NULL )
	return;
    if( !cfg_read )
    {
	q_config();
	++cfg_read;
    }
    fin = fopen(echolog_fname, "r");
    if( fin == NULL )
	return;
    while( !feof(fin) )
    {
	AREA *	ar;

	fgets(buf, MAX_TAGLEN, fin);
	if( strchr(buf, '\n') )
	    *strchr(buf, '\n') = 0;
	ar = findarea(buf);
	if( ar )
	    ar->dolink = 0;
    }
    fclose(fin);
    if( kill )
	remove(echolog_fname);
}




PRIVATE void
help(void)
{
    printf("\n\n"
	   "Usage: QQLink [-c<Squish.CFG>] [<Areas>] [-switches] [-f<EchoToss.Log>]\n");
    printf("   or: QQLink [-a<Areas.BBS>] [<Areas>] [-switches] [-f<EchoToss.Log>]\n");
    printf("\n"
	   "    -q[n=1] suppress some screen display (0: all, 1: states, 2: percent)\n"
	   "    -b[n=2]file write messages to logfile (n has same meaning as in -q)\n");
    printf("    -P<n> use <n> threads to link several areas in parallel (default %u)\n",
	   NUM_THREADS);
    printf("    -c, -g, -M3 and -a are ways of specifying your areas.\n"
	   "    -f names a file containing a list of area tags to link.\n"
	   "    -r means relink any already linked messages.\n"
	   "    -l means link every area listed in Squish.CFG or Areas.BBS\n"
	   "    -n means remove old linking information before relink\n"
	   "    -pt links passthru areas if asked to\n"
	   "    -s, -m, -t and -T select to do more than just FTS9 linking\n"
	   "    -S link without using a .QQL file (slow!)\n"
	   "    -k means delete <EchoToss.log> after linking.\n"
	   "    -x names a file containing a list of area tags NOT to link\n"
	   "    -o file where list of area actually linked will be written to\n");
    printf("    <Areas> is a list of echo tags of areas you want linked.\n");
    printf("\n"
	   "Example: QQLink -c\\SQUISH\\SQUISH.CFG NETMAIL -fIN.LOG\n"
	   "         This will read \\SQUISH\\SQUISH.CFG for area definitions, and link\n"
	   "         any new messages in NETMAIL and the areas listed in IN.LOG.\n");
    exit(1);
}




PRIVATE void
write_outlog(const char *fn)
{
    AREA *	ar;
    FILE * 	fout = fopen(fn," a");

    if( fout == NULL )
	return;
    for( ar = rootarea; ar ; ar = ar->next )
	if( ar->dolink )
	    fprintf(fout, "%s\n", ar->tag);
    fclose(fout);
}




PRIVATE void
pre_param(char const *arg)
{
    if( !PARJCMP(arg, "-q") )
    {
	if( *(arg+2) == '\0' )
	    SetDisplaylevel(1);
	else
	    SetDisplaylevel(atoi(arg+2));
    }
    else if( !PARJCMP(arg, "-b") )
    {
	HFILE	h;
	ULONG	ul;
	int	off = 2;

	if( isdigit(*(arg+2)) )
	{
	    SetLoglevel(atoi(arg+2));
	    off = 3;
	}
	DosOpen((PVOID)(arg+off), &h, &ul, 0, FILE_NORMAL,
		OPEN_ACTION_CREATE_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS,
		OPEN_FLAGS_FAIL_ON_ERROR|OPEN_FLAGS_NO_CACHE
		|OPEN_FLAGS_SEQUENTIAL|OPEN_SHARE_DENYWRITE|OPEN_ACCESS_WRITEONLY,
		NULL);
	DosSetFilePtr(h, 0, FILE_END, &ul);
	SetLogfile(h);
    }
    else if( !PARJCMP(arg, "-P") )
    {
	if( *(arg+2) == '\0' )
	    cThreads = NUM_THREADS;
	else
	    cThreads = atoi(arg+2);
    }
    return;
}




PRIVATE void
post_param(const char *arg)
{
    if( !stricmp(arg, "/?") )
	help();
    else if( !strcmp(arg,"-?") )
	help();
    else if( !PARJCMP(arg,"-q") )
	    ;
    else if( !PARJCMP(arg, "-P") )
	    ;
    else if( !PARJCMP(arg,"-b") )
	    ;
    else if( !strcmp(arg,"-k") )
	kill_log = 1;
    else if( !strcmp(arg,"-k+") )
	kill_log = 1;
    else if( !strcmp(arg,"-k-") )
	kill_log = 0;
    else if( !strcmp(arg,"-r") )
	kill_qql = 1;
    else if( !strcmp(arg,"-l") )
	link_all = 1;
    else if( !strcmp(arg,"-s") )
	link_meth = LINK_SUBJ;
    else if( !strcmp(arg,"-m") )
	link_meth = LINK_COM1;
    else if( !strcmp(arg,"-n") )
	link_clean = 1;
    else if( !strcmp(arg,"-p") )
	link_prune = 1;
    else if( !strcmp(arg,"-pt") )
	link_passthru = 1;
    else if( !strcmp(arg,"-S") )
	link_slow = 1;
    else if( !PARJCMP(arg,"-f") )
	read_echolog(arg+2, kill_log);
    else if( !PARJCMP(arg,"-x") )
	read_antilog(arg+2, kill_log);
    else if( !PARJCMP(arg,"-o") )
	outlog_fname = strdup(arg+2);
    else if( !PARJCMP(arg,"-c") )
	squishcfg_fname = strdup(arg+2);
    else if( !PARJCMP(arg,"-a") )
	areasbbs_fname = strdup(arg+2);
    else if(!PARJCMP(arg,"-M3"))
	max3areadat_fname = strdup(arg+3);
    else if(!PARJCMP(arg,"-g"))
	gepath = strdup(arg+2);
    else if( !PARJCMP(arg,"-T") )
	trunc_qq = atoi(arg+2);
    else if( !PARJCMP(arg,"-t") )
	all_name = strdup(arg+2);
#if 0
    else if( !PARCMP(arg,"-d") )
	cancel_parm=arg+2;
#endif
#if 0
    else if( !PARCMP(arg,"-*") )
	/* add arg+2 as a *.MSG area */;
#endif
#if 0
    else if( !PARCMP(arg,"-$") )
	/* add arg+2 as a *.SQ? area */;
#endif
    else if( !stricmp(arg,"/HELP") )
	help();
    else if( !stricmp(arg,"/KILL") )
	kill_log = 1;
    else if( !stricmp(arg,"/NOKILL") )
	kill_log = 0;
    else if( !stricmp(arg,"/RELINK") )
	kill_qql = 1;
    else if( !stricmp(arg,"/ALL") )
	link_all = 1;
    else if( !stricmp(arg,"/SUBJ") )
	link_meth = LINK_SUBJ;
    else if( !stricmp(arg,"/MIXED") )
	link_meth = LINK_COM1;
    else if( !stricmp(arg,"/CLEAN") )
	link_clean = 1;
    else if( !stricmp(arg,"/PRUNE") )
	link_prune = 1;
    else if( !stricmp(arg,"/LINKPASS") )
	link_passthru = 1;
    else if( !stricmp(arg,"/SLOW") )
	link_slow = 1;
    else if( !PARCMP(arg,"/LOG:") )
	read_echolog(arg+5, kill_log);
    else if( !PARCMP(arg,"/XLOG:") )
	read_antilog(arg+6, kill_log);
    else if( !PARCMP(arg,"/OUT:") )
	outlog_fname = strdup(arg+5);
    else if( !PARCMP(arg,"/CFG:") )
	squishcfg_fname = strdup(arg+5);
    else if( !PARCMP(arg,"/AREAS:") )
	areasbbs_fname = strdup(arg+7);
    else if(!PARCMP(arg,"/MAX3:"))
	max3areadat_fname = strdup(arg+6);
    else if(!PARCMP(arg,"/GECHO:"))
	gepath = strdup(arg+7);
    else if( !PARCMP(arg,"/TRUNC:") )
	trunc_qq = atoi(arg+7);
    else if( !PARCMP(arg,"/TOALL:") )
	all_name = strdup(arg+7);
    else if( !stricmp(arg,"/TOALL") )
	all_name = NULL;
#if 0
    else if( !PARCMP(arg,"-d") )
	cancel_parm = arg+2;
#endif
#if 0
    else if( !PARCMP(arg,"/MSG:") )
	/* add arg+2 as a *.MSG area */;
#endif
#if 0
    else if( !PARCMP(arg,"/SQ:") )
	/* add arg+2 as a *.SQ? area */;
#endif
    else if( *arg == '/'  ||  *arg == '-' )
    {
	Verbose(0, "QQLINK",
		"Unrecognized option: \"%s\"."
		"\n(Run QQLink -? for a list of options)", arg);
	exit(7);
    }
}




PRIVATE void
env_args(char *buf,void (*do_arg)(const char *arg))
{
    char *	s = buf;

    while( *s )
    {
	if( isspace(*s) )
	    ++s;
	else
	{
	    char *	end = s + strcspn(s, " \r\n\x9");
	    char	old = *end;

	    *end = 0;
	    do_arg(s);
	    *(s=end) = old;
	}
    }
}




int
main(int argc,char **argv)
{
    struct _minf minf = {0, 0, 0};
    UINT	i, alinked = 0;
    ULONG	nlinked = 0;
    time_t	tstart, tdone;
    APIRET	rc;


    SetDisplaylevel(2);
    SetLoglevel(2);
    if( getenv("QQLINK") )
	env_args(getenv("QQLINK"), pre_param);
    for( i = 1; i < argc; ++i)
	pre_param(argv[i]);

    Verbose(2, "QQLINK", "QQLink " VERSION_STR " for Squish(tm)  *  A Strange Software");
    Verbose(2, "QQLINK", "Copyright (c) 1993 Josh Parsons (3:771/330.64). All rights reserved.");
#ifdef PORTED_BY
    Verbose(2, "QQLINK", "Ported to %s by %s.", PORTED_TO, PORTED_BY);
#endif
#ifdef MODIFIED_BY
    Verbose(2, "QQLINK", "Modifications by %s.", MODIFIED_BY);
#endif

    if( getenv("QQLINK") )
	env_args(getenv("QQLINK"), post_param);
    for( i = 1; i < argc; ++i )
	post_param(argv[i]);
    if( !cfg_read )
    {
	q_config();
	++cfg_read;
    }

    MsgOpenApi(&minf);
    time(&tstart);
    for( i = 1; i < argc; ++i)
	if( *argv[i] != '-'  &&  *argv[i] != '/' )
	{
	    AREA *	ar;

	    if( *argv[i] == '*' )
		add_area(argv[i], argv[i]+1, 0, 0);
	    else if( *argv[i] == '$' )
		add_area(argv[i], argv[i]+1, 1, 0);
	    ar = findarea(argv[i]);
	    if( ar == NULL )
	    {
		Verbose(0, "Unrecognised area \"%s\"", argv[i]);
		exit(5);
	    }
	    ar->dolink = 1;
	}

    /* After we've found all areas to be linked
     * do the actual linking. */

    rc = DoLinking(&alinked, &nlinked);
    if( rc != 0 )
	return rc;				/* immediate exit! */


    time(&tdone);
    tdone -= tstart;
    if( tdone == 0 )
	tdone = 1;

    MsgCloseApi();


    /* Display some nice-to-know numbers. */

    if( alinked )
    {
	Verbose(1, "QQLINK",
		"%lu message%s in %d area%s linked in %ld second%s (%d.%d msgs/sec)",
		nlinked, nlinked == 1 ? "" : "s",
		alinked, alinked == 1 ? "" : "s",
		tdone, tdone == 1 ? "" : "s",
		(int)(nlinked/tdone),
		(int)(((10L*nlinked) / tdone)%10L));
    }
    else
    {
	Verbose(0, "QQLINK", "No areas linked.");
	if( argc == 1 )
	    Verbose(0, "QQLINK", "(Run QQLink -? for a list of options)");
    }


    /* Cleanup */

    if( outlog_fname )
	write_outlog(outlog_fname);
    return 0;
}
