/*
 * $Source: r:/source/msgapi/qqlink/RCS/testid.c,v $
 * $Revision: 1.2 $
 * $Date: 1998/09/02 01:32:06 $
 * $Locker:  $
 *
 *	Test ID calculation.
 *
 * $Log: testid.c,v $
 * Revision 1.2  1998/09/02 01:32:06  vitus
 * - added alternate ID calculations: MD5, CRC32 (used)
 *
 * Revision 1.1  1998/09/01 22:59:17  vitus
 * Initial revision
 */
static char const vcid[]="$Id: testid.c,v 1.2 1998/09/02 01:32:06 vitus Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MAKEID_METHOD	2			/* 0: QQLink, 1: MD5, 2:CRC32 */


#if MAKEID_METHOD == 1
# include "md5.h"
#endif
#if MAKEID_METHOD == 2
# include "crc32.h"
#endif


#define PRIVATE

typedef long		MSGID;
typedef MSGID *		PMSGID;


#define MIN(a,b)	((a)<(b)?(a):(b))


int	trunc_qq = 0;


#if MAKEID_METHOD == 0
PRIVATE void
makeid(MSGID *id,char const *s)
{
    char	b = 0;
    int		ctr = trunc_qq;

    while( *s )
    {
	(*id) += ((long)tolower(*s)) << b;
	b = (b+1) % 24;
	++s;
	if( (--ctr) == 0 )
	    break;
    }
    if( *id == 0 )
	*id = 1;		/* ID=0 is used for "no ID found" by getid() */
}
#endif


#if MAKEID_METHOD == 1
PRIVATE void
makeid(PMSGID id,const char *s)
{
    char const *	buf = strlwr(strdup(s));
    int const		size = MIN(strlen(buf),trunc_qq);
    struct MD5Context	ctx;
    unsigned char	digest[16];
    unsigned long const * p = (unsigned long const *)digest;

    MD5Init(&ctx);
    MD5Update(&ctx, buf, size);
    MD5Final(digest, &ctx);
    *id = (((p[0] + p[1]) >> 1)
	   + ((p[2] + p[3]) >> 1));

    if( *id == 0 )
	*id = 1;		/* ID=0 is used for "no ID found" by getid() */
    return;
}
#endif


#if MAKEID_METHOD == 2
PRIVATE void
makeid(PMSGID id,const char *s)
{
    int		ctr = trunc_qq;
    CRC32	crc;        

    crc32Init(&crc);
    while( *s )
    {
	crc32Update(&crc, tolower(*s));
 	++s;
 	if( (--ctr) == 0 )
	    break;
    }
    *id = crc32Value(&crc);

    if( *id == 0 )
	*id = 1;		/* ID=0 is used for "no ID found" by getid() */
    return;
}
#endif




int
main(int argc,char *argv[])
{
    MSGID	id;

    makeid(&id, argv[1]);
    printf("\'%s' gives %#lx\n",argv[1], id);
    return MAKEID_METHOD;
}

